//
// GameX - Input-related WindowsDX Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


#include "gamex-win-dx.hpp"




bool WindowsDX::UpdateInput (void)
{
	bool skip_keyboard = (di_keyboard == NULL);
	bool skip_mouse = (di_mouse == NULL) || win_request_no_mouse_input;

	if(!skip_keyboard) {
		if(FAILED(di_keyboard->GetDeviceState(256, (LPVOID)di_keystate))) {
			if(SUCCEEDED(di_keyboard->Acquire())) {
				if(FAILED(di_keyboard->GetDeviceState(256, (LPVOID)di_keystate))) {
					skip_keyboard = true;
				}
			}
		}

		if(!skip_keyboard)
		{
			{
				for(KeyID k = 0 ; k < 255 ; k++) {
					if(di_keypressed[k] & INPUT_PRESSED) // if just pressed last time,
						di_keypressed[k] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}

			// fix windows oddity that makes the right shift button register as unpressed
			// if an arrow key is pressed while it's down
			if(!IsKeyDown(KEY_RSHIFT) && (GetAsyncKeyState(VK_RSHIFT) & 0x8000))
				di_keystate[KEY_RSHIFT] |= INPUT_PRESSED;

			// transfer/record unique (non-held) keypresses
			for(KeyID k = 0 ; k < 255 ; k++) {
				if(di_keypressed[k] & INPUT_HELD) {
					if(!di_keystate[k])
						di_keypressed[k] = di_keystate[k];
				} else {
					di_keypressed[k] = di_keystate[k];
				}
			}

			if(IsEscapeReserved() && GetKeyPress(KEY_ESCAPE) // ESC is reserved to quit unless RUN_USEESCAPEKEY is on
			|| IsCtrlDown() && GetKeyPress(KEY_Q) // whenever CTRL is down, Q is reserved to quit (nonstandard but intuitive)
			|| IsAltDown() && GetKeyPress(KEY_F4)) // whenever ALT is down, F4 is reserved to quit (faster than in WindowFunc)
				Quit();


			if(IsAltEnterReserved() && GetKeyPress(KEY_ENTER) && IsAltDown()) // ALT+ENTER is reserved to toggle fullscreen mode unless RUN_NOALTENTERTOGGLE is on
				GameX.ToggleFullscreen();
		}
	}

	if(!skip_mouse) {

		if(FAILED(di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)&di_mousestate))) {
			if(SUCCEEDED(di_mouse->Acquire())) {
				if(FAILED(di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)&di_mousestate))) {
					skip_mouse = true;
				}
			}
		}

		if(!skip_mouse) {

			{
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					if(di_mousepressed[b] & INPUT_PRESSED) // if just pressed last time,
						di_mousepressed[b] ^= (INPUT_HELD|INPUT_PRESSED); // switch pressed off, held on
				}
			}
			
			// if the mouse isn't outside of the game,
			if(win_fullscreen || (mouse_x>0 && mouse_y>0 && GetMouseX()<win_width && GetMouseY()<win_height)) {
				// transfer/record unique (non-held) mousepresses
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					if(!di_mousestate.rgbButtons[b])
						mouse_down[b] = 0;
					else if(!mouse_down[b])
		/// 			di_mousestate.rgbButtons[b] = 0; // seems to cause mouse click detection to lag after typing
						;
					else if(mouse_down[b] == 2 && di_mousestate.rgbButtons[b])
						di_mousestate.rgbButtons[b] |= 0x01;
					if(di_mousepressed[b] & INPUT_HELD) {
						if(!di_mousestate.rgbButtons[b])
							di_mousepressed[b] = di_keystate[b];
					} else {
						di_mousepressed[b] = di_mousestate.rgbButtons[b];
					}
				}
			}
			if(di_mousestop) {
				di_mousestate.lX = di_mousestate.lY = di_mousestate.lZ = 0;
				if(win_active)
					di_mousestop--;

				// unregister all clicks at this time (so the game doesn't respond to window-activating clicks)
				for(MouseButtonID b = 0 ; b < 4 ; b++) {
					mouse_down[b] = 0;
					di_mousepressed[b] = 0;
				}
			}

			POINT mousePoint;
			GetCursorPos(&mousePoint);
			mousePoint.x -= win_xpos+win_xoffset;
			mousePoint.y -= win_ypos+win_yoffset;
			SetMouseMove(mousePoint.x,mousePoint.y);
		}
	}

	return true;
}

bool WindowsDX::AnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(GetKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}

bool WindowsDX::PeekAnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(PeekKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}


void WindowsDX::GetMouseButtonString (char * output, MouseButtonID button)
{
	switch(button)
	{
	case MOUSE_LEFT: strcpy(output,"LClick"); break;
	case MOUSE_RIGHT: strcpy(output,"RClick"); break;
	case MOUSE_MIDDLE: strcpy(output,"MClick"); break;
	case MOUSE_EXTRA: strcpy(output,"EClick"); break;
	default: strcpy(output,"unknown"); break;
	}
}



void WindowsDX::GetKeyString (char * output, KeyID key)
{
	switch(key)
	{
	case KEY_SPACE: strcpy(output,"SpaceBar"); break;
	case KEY_ENTER: strcpy(output,"Enter"); break;
	case KEY_A: strcpy(output,"A"); break;
	case KEY_B: strcpy(output,"B"); break;
	case KEY_C: strcpy(output,"C"); break;
	case KEY_D: strcpy(output,"D"); break;
	case KEY_E: strcpy(output,"E"); break;
	case KEY_F: strcpy(output,"F"); break;
	case KEY_G: strcpy(output,"G"); break;
	case KEY_H: strcpy(output,"H"); break;
	case KEY_I: strcpy(output,"I"); break;
	case KEY_J: strcpy(output,"J"); break;
	case KEY_K: strcpy(output,"K"); break;
	case KEY_L: strcpy(output,"L"); break;
	case KEY_M: strcpy(output,"M"); break;
	case KEY_N: strcpy(output,"N"); break;
	case KEY_O: strcpy(output,"O"); break;
	case KEY_P: strcpy(output,"P"); break;
	case KEY_Q: strcpy(output,"Q"); break;
	case KEY_R: strcpy(output,"R"); break;
	case KEY_S: strcpy(output,"S"); break;
	case KEY_T: strcpy(output,"T"); break;
	case KEY_U: strcpy(output,"U"); break;
	case KEY_V: strcpy(output,"V"); break;
	case KEY_W: strcpy(output,"W"); break;
	case KEY_X: strcpy(output,"X"); break;
	case KEY_Y: strcpy(output,"Y"); break;
	case KEY_Z: strcpy(output,"Z"); break;
	case KEY_0: strcpy(output,"0"); break;
	case KEY_1: strcpy(output,"1"); break;
	case KEY_2: strcpy(output,"2"); break;
	case KEY_3: strcpy(output,"3"); break;
	case KEY_4: strcpy(output,"4"); break;
	case KEY_5: strcpy(output,"5"); break;
	case KEY_6: strcpy(output,"6"); break;
	case KEY_7: strcpy(output,"7"); break;
	case KEY_8: strcpy(output,"8"); break;
	case KEY_9: strcpy(output,"9"); break;
	case KEY_TILDE: strcpy(output,"`"); break;
	case KEY_APOSTROPHE: strcpy(output,"'"); break;
	case KEY_SEMICOLON: strcpy(output,";"); break;
	case KEY_LBRACKET: strcpy(output,"["); break;
	case KEY_RBRACKET: strcpy(output,"]"); break;
	case KEY_COMMA: strcpy(output,","); break;
	case KEY_PERIOD: strcpy(output,"."); break;
	case KEY_NUM0: strcpy(output,"num0"); break;
	case KEY_NUM1: strcpy(output,"num1"); break;
	case KEY_NUM2: strcpy(output,"num2"); break;
	case KEY_NUM3: strcpy(output,"num3"); break;
	case KEY_NUM4: strcpy(output,"num4"); break;
	case KEY_NUM5: strcpy(output,"num5"); break;
	case KEY_NUM6: strcpy(output,"num6"); break;
	case KEY_NUM7: strcpy(output,"num7"); break;
	case KEY_NUM8: strcpy(output,"num8"); break;
	case KEY_NUM9: strcpy(output,"num9"); break;
	case KEY_NUMTIMES: strcpy(output,"num*"); break;
	case KEY_NUMDIVIDE: strcpy(output,"num/"); break;
	case KEY_NUMPLUS: strcpy(output,"num+"); break;
	case KEY_NUMMINUS: strcpy(output,"num-"); break;
	case KEY_NUMDOT: strcpy(output,"NUM."); break;
	case KEY_LEFT: strcpy(output,"LeftArrow"); break;
	case KEY_UP: strcpy(output,"UpArrow"); break;
	case KEY_RIGHT: strcpy(output,"RightArrow"); break;
	case KEY_DOWN: strcpy(output,"DownArrow"); break;
	case KEY_CAPS: strcpy(output,"CapsLock"); break;
	case KEY_TAB: strcpy(output,"Tab"); break;
	case KEY_ESCAPE: strcpy(output,"Escape"); break;
	case KEY_LSHIFT: strcpy(output,"LShift"); break;
	case KEY_RSHIFT: strcpy(output,"RShift"); break;
	case KEY_LALT: strcpy(output,"LAlt"); break;
	case KEY_RALT: strcpy(output,"RAlt"); break;
	case KEY_LWIN: strcpy(output,"LWin"); break;
	case KEY_RWIN: strcpy(output,"RWin"); break;
	case KEY_LCTRL: strcpy(output,"LCtrl"); break;
	case KEY_RCTRL: strcpy(output,"RCtrl"); break;
	case KEY_BACKSLASH: strcpy(output,"/"); break;
	case KEY_FORWARDSLASH: strcpy(output,"\\"); break;
	case KEY_MINUS: strcpy(output,"-"); break;
	case KEY_PLUS: strcpy(output,"="); break;
	case KEY_BACKSPACE: strcpy(output,"Backspace"); break;
	case KEY_DELETE: strcpy(output,"Delete"); break;
	case KEY_HOME: strcpy(output,"Home"); break;
	case KEY_END: strcpy(output,"End"); break;
	case KEY_INSERT: strcpy(output,"Insert"); break;
	case KEY_PAGEUP: strcpy(output,"PageUp"); break;
	case KEY_PAGEDOWN: strcpy(output,"PageDown"); break;
	case KEY_PAUSE: strcpy(output,"Pause"); break;
	case KEY_NUMLOCK: strcpy(output,"NumLock"); break;
	case KEY_PRINT: strcpy(output,"PrintScreen"); break;
	case KEY_F1: strcpy(output,"F1"); break;
	case KEY_F2: strcpy(output,"F2"); break;
	case KEY_F3: strcpy(output,"F3"); break;
	case KEY_F4: strcpy(output,"F4"); break;
	case KEY_F5: strcpy(output,"F5"); break;
	case KEY_F6: strcpy(output,"F6"); break;
	case KEY_F7: strcpy(output,"F7"); break;
	case KEY_F8: strcpy(output,"F8"); break;
	case KEY_F9: strcpy(output,"F9"); break;
	case KEY_F10: strcpy(output,"F10"); break;
	case KEY_F11: strcpy(output,"F11"); break;
	case KEY_F12: strcpy(output,"F12"); break;
	default: strcpy(output,"unknown"); break;
	}
}




int WindowsDX::GetKeyDownEvent (void)
{
	if(!win_keydownbuf.empty())
	{
		int ret = win_keydownbuf.front();
		win_keydownbuf.pop();
		return ret;
	}
	return 0;
}
int WindowsDX::GetKeyUpEvent (void)
{
	if(!win_keyupbuf.empty())
	{
		int ret = win_keyupbuf.front();
		win_keyupbuf.pop();
		return ret;
	}
	return 0;
}

int WindowsDX::PeekKeyDownEvent (void)
{
	if(!win_keydownbuf.empty())
		return win_keydownbuf.front();
	return 0;
}

int WindowsDX::PeekKeyUpEvent (void)
{
	if(!win_keyupbuf.empty())
		return win_keyupbuf.front();
	return 0;
}

void WindowsDX::ClearKeyDownEvents (void)
{
	while(!win_keydownbuf.empty())
		win_keydownbuf.pop();
}

void WindowsDX::ClearKeyUpEvents (void)
{
	while(!win_keyupbuf.empty())
		win_keyupbuf.pop();
}

MouseButtonID WindowsDX::GetMouseDownEvent (void)
{
	if(!win_mousedownbuf.empty())
	{
		int ret = win_mousedownbuf.front();
		win_mousedownbuf.pop();
		return ret;
	}
	return MOUSE_NOTHING;
}
MouseButtonID WindowsDX::GetMouseUpEvent (void)
{
	if(!win_mouseupbuf.empty())
	{
		int ret = win_mouseupbuf.front();
		win_mouseupbuf.pop();
		return ret;
	}
	return MOUSE_NOTHING;
}
MouseButtonID WindowsDX::PeekMouseDownEvent (void)
{
	if(!win_mousedownbuf.empty())
		return win_mousedownbuf.front();
	return MOUSE_NOTHING;
}
MouseButtonID WindowsDX::PeekMouseUpEvent (void)
{
	if(!win_mouseupbuf.empty())
		return win_mouseupbuf.front();
	return MOUSE_NOTHING;
}
void WindowsDX::ClearMouseDownEvents (void)
{
	while(!win_mousedownbuf.empty())
		win_mousedownbuf.pop();
}
void WindowsDX::ClearMouseUpEvents (void)
{
	while(!win_mouseupbuf.empty())
		win_mouseupbuf.pop();
}

char WindowsDX::GetBufferedKeyPress (void)
{
	if(!win_keycharbuf.empty())
	{
		int ret = win_keycharbuf.front();
		win_keycharbuf.pop();
		return ret;
	}
	return '\0';
}

char WindowsDX::PeekBufferedKeyPress (void)
{
	if(!win_keycharbuf.empty())
		return win_keycharbuf.front();
	return '\0';
}

void WindowsDX::ClearKeyBuffer (void)
{
	while(!win_keycharbuf.empty())
		win_keycharbuf.front();
}

void WindowsDX::AddBufferedKeyPress (char ch)
{
	if(win_keycharbuf.size() < WINDX_BUFSIZE)
		win_keycharbuf.push(ch);
}
void WindowsDX::AddKeyDownEvent (KeyID key)
{
	if(win_keydownbuf.size() < WINDX_BUFSIZE)
		win_keydownbuf.push(key);
}
void WindowsDX::AddKeyUpEvent (KeyID key)
{
	if(win_keyupbuf.size() < WINDX_BUFSIZE)
		win_keyupbuf.push(key);
}
void WindowsDX::AddMouseDownEvent (MouseButtonID btn)
{
	if(win_mousedownbuf.size() < WINDX_BUFSIZE)
		win_mousedownbuf.push(btn);
}
void WindowsDX::AddMouseUpEvent (MouseButtonID btn)
{
	if(win_mouseupbuf.size() < WINDX_BUFSIZE)
		win_mouseupbuf.push(btn);
}


